/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.document.bescheid.binaryfile;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.apache.commons.io.IOUtils;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.stereotype.Service;

import com.google.protobuf.ByteString;

import de.ozgcloud.common.binaryfile.FileId;
import de.ozgcloud.common.binaryfile.GrpcFileUploadUtils;
import de.ozgcloud.common.binaryfile.GrpcFileUploadUtils.FileSender;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.document.DocumentManagerConfiguration;
import de.ozgcloud.document.bescheid.BescheidCallContextAttachingInterceptor;
import de.ozgcloud.document.bescheid.BescheidResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceStub;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileMetaData;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileResponse;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import net.devh.boot.grpc.client.inject.GrpcClient;

@Service
@RequiredArgsConstructor
class BinaryFileRemoteService {

	static final String VORGANG_ATTACHMENT_FIELD = "bescheid";

	@GrpcClient(DocumentManagerConfiguration.GRPC_VORGANG_MANAGER_NAME)
	private final BinaryFileServiceStub binaryFileRemoteStub;
	@Qualifier(DocumentManagerConfiguration.CALL_CONTEXT_CLIENT_INTERCEPTOR_NAME)
	private final BescheidCallContextAttachingInterceptor callContextInterceptor;

	public FileId uploadBescheidFile(@NonNull BescheidResponse bescheid) {
		try (var in = openFile(bescheid.getBescheidFile())) {
			var resultFuture = GrpcFileUploadUtils.createSender(this::buildChunkRequest, in,
							this::buildCallStreamObserver)
					.withMetaData(buildMetaDataRequest(bescheid))
					.send();

			return FileId.from(waitUntilFutureToComplete(resultFuture, in).getFileId());
		} catch (IOException e) {
			throw new TechnicalException("Error on uploading file.", e);
		}
	}

	InputStream openFile(File file) {
		try {
			return new FileInputStream(file);
		} catch (FileNotFoundException e) {
			throw new TechnicalException("File to upload not found.", e);
		}
	}

	GrpcUploadBinaryFileRequest buildMetaDataRequest(BescheidResponse bescheid) {
		return GrpcUploadBinaryFileRequest.newBuilder()
				.setMetadata(GrpcUploadBinaryFileMetaData.newBuilder()
						.setVorgangId(bescheid.getVorgangId().toString())
						.setField(VORGANG_ATTACHMENT_FIELD)
						.setContentType(bescheid.getContentType())
						.setSize(bescheid.getBescheidFile().length())
						.setFileName(bescheid.getBescheidFileName())
						.build())
				.build();
	}

	GrpcUploadBinaryFileRequest buildChunkRequest(byte[] bytes, Integer length) {
		return GrpcUploadBinaryFileRequest.newBuilder().setFileContent((ByteString.copyFrom(bytes, 0, length))).build();
	}

	CallStreamObserver<GrpcUploadBinaryFileRequest> buildCallStreamObserver(
			StreamObserver<GrpcUploadBinaryFileResponse> responseObserver) {
		return (CallStreamObserver<GrpcUploadBinaryFileRequest>) binaryFileRemoteStub.withInterceptors(callContextInterceptor)
				.uploadBinaryFileAsStream(responseObserver);
	}

	GrpcUploadBinaryFileResponse waitUntilFutureToComplete(FileSender<GrpcUploadBinaryFileRequest, GrpcUploadBinaryFileResponse> fileSender,
			InputStream fileContentStream) {
		try {
			return fileSender.getResultFuture().get(10, TimeUnit.MINUTES);
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			fileSender.cancelOnError(e);
			throw new TechnicalException("Waiting for finishing upload was interrupted.", e);
		} catch (ExecutionException | TimeoutException e) {
			fileSender.cancelOnTimeout();
			throw new TechnicalException("Error / Timeout on uploading data.", e);
		} finally {
			IOUtils.closeQuietly(fileContentStream);
		}
	}
}
