/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.document;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.web.server.ResponseStatusException;

import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.common.command.grpc.CommandMapper;
import de.ozgcloud.apilib.common.datatypes.GenericId;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangId;
import de.ozgcloud.document.bescheid.Bescheid;
import de.ozgcloud.document.bescheid.BescheidCallContextAttachingInterceptor;
import de.ozgcloud.document.bescheid.BescheidTestFactory;
import de.ozgcloud.document.bescheid.BescheidResponse;
import de.ozgcloud.document.bescheid.BescheidResponseTestFactory;
import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandTestFactory;
import de.ozgcloud.document.common.attached_item.AttachedItem;
import de.ozgcloud.document.common.attached_item.AttachedItemService;
import de.ozgcloud.document.common.attached_item.AttachedItemTestFactory;
import de.ozgcloud.common.binaryfile.FileId;
import de.ozgcloud.common.errorhandling.TechnicalException;

class DocumentServiceTest {

	private static final String FILE_ID = "file-id";
	private static final String RELATION_ID = "relation-id";

	@Spy
	@InjectMocks
	private DocumentService service;

	@Mock
	private OzgCloudCommandService commandService;
	@Mock
	private CommandMapper commandMapper;
	@Mock
	private AttachedItemService attachedItemService;
	@Mock
	private DocumentMapper documentMapper;

	private final Command command = CommandTestFactory.createBuilder().relationId(RELATION_ID).build();

	@Nested
	class TestCreateBescheidDocument {

		private static final String CREATED_ATTACHED_ITEM_ID = "document-id";

		@Mock
		private OzgCloudCommand ozgCommand;

		@Nested
		class TestWithCommand {

			private final Command command = CommandTestFactory.createBuilder().relationId(RELATION_ID)
					.bodyObject(Map.of(DocumentService.FIELD_DOCUMENT_FILE, FILE_ID)).build();

			@BeforeEach
			void init() {
				doReturn(CREATED_ATTACHED_ITEM_ID).when(service).createBescheidDocument(any(), anyMap());
			}

			@Test
			void shouldCallBuildItemMap() {
				service.createBescheidDocument(command);

				verify(service).buildItemMap(command);
			}

			@Test
			void shouldCallCreateBescheidDocument() {
				var itemMap = Map.of("key", (Object) "value");
				doReturn(itemMap).when(service).buildItemMap(any(Command.class));

				service.createBescheidDocument(command);

				verify(service).createBescheidDocument(command, itemMap);
			}

			@Test
			void shouldReturnDocumentId() {
				String result = service.createBescheidDocument(command);

				assertThat(result).isEqualTo(CREATED_ATTACHED_ITEM_ID);
			}
		}

		@Nested
		class TestWithBescheid {

			private final BescheidResponse bescheid = BescheidResponseTestFactory.create()
					.withBescheidFileId(FileId.from(DocumentTestFactory.ID));

			@BeforeEach
			void init() {
				doReturn(CREATED_ATTACHED_ITEM_ID).when(service).createBescheidDocument(any(), anyMap());
			}

			@Test
			void shouldCallBuildItemMap() {
				service.createBescheidDocument(command, bescheid);

				verify(service).buildItemMap(bescheid);
			}

			@Test
			void shouldCallCreateBescheidDocument() {
				var itemMap = Map.of("key", (Object) "value");
				doReturn(itemMap).when(service).buildItemMap(any(BescheidResponse.class));

				service.createBescheidDocument(command, bescheid);

				verify(service).createBescheidDocument(command, itemMap);
			}

			@Test
			void shouldReturnDocumentId() {
				String result = service.createBescheidDocument(command, bescheid);

				assertThat(result).isEqualTo(CREATED_ATTACHED_ITEM_ID);
			}
		}

		@Nested
		class TestWithCommandAndItemMap {

			private static final Map<String, Object> BODY_OBJECT = Map.of(DocumentService.FIELD_DOCUMENT_FILE, FILE_ID);

			@BeforeEach
			void init() {
				when(commandService.createAndWaitUntilDone(any())).thenReturn(ozgCommand);
				doNothing().when(service).validateBescheidItem(any());
			}

			@Test
			void shouldCallValidateBescheidItem() {
				service.createBescheidDocument(command, BODY_OBJECT);

				verify(service).validateBescheidItem(RELATION_ID);
			}

			@Test
			void shouldCallCommandService() {
				doReturn(ozgCommand).when(service).buildCreateDocumentOzgCommand(any(), anyMap());

				service.createBescheidDocument(command, BODY_OBJECT);

				verify(commandService).createAndWaitUntilDone(ozgCommand);
			}

			@Test
			void shouldReturnDocumentId() {
				when(ozgCommand.getCreatedResource()).thenReturn(CREATED_ATTACHED_ITEM_ID);

				String documentId = service.createBescheidDocument(command, BODY_OBJECT);

				assertThat(documentId).isEqualTo(CREATED_ATTACHED_ITEM_ID);
			}
		}
	}

	@Nested
	class TestValidateBescheidItem {

		@Test
		void shouldCallGetBescheidItem() {
			when(attachedItemService.getItem(any())).thenReturn(AttachedItemTestFactory.createBescheid());

			validateBescheidItem();

			verify(attachedItemService).getItem(AttachedItemTestFactory.ID);
		}

		@Test
		void shouldThrowExceptionWhenNoBescheidDraft() {
			when(attachedItemService.getItem(any())).thenThrow(ResponseStatusException.class);

			assertThrows(ResponseStatusException.class, this::validateBescheidItem);
		}

		@Test
		void shouldNotThrowExceptionIfNotDraft() {
			when(attachedItemService.getItem(any())).thenReturn(
					AttachedItemTestFactory.createBescheidBuilder().clearItem().itemEntry(Bescheid.FIELD_STATUS, "not-draft").build());

			assertThrows(TechnicalException.class, this::validateBescheidItem);
		}

		@Test
		void shouldValidateBescheidItem() {
			when(attachedItemService.getItem(any())).thenReturn(AttachedItemTestFactory.createBescheid());

			assertDoesNotThrow(this::validateBescheidItem);
		}

		void validateBescheidItem() {
			service.validateBescheidItem(AttachedItemTestFactory.ID);
		}
	}

	@Nested
	class TestBuildCreateDocumentOzgCommand {

		private static final Map<String, Object> BODY_OBJECT = Map.of(DocumentService.FIELD_DOCUMENT_FILE, FILE_ID);

		@Test
		void shouldSetOrder() {
			var ozgCommand = service.buildCreateDocumentOzgCommand(command, BODY_OBJECT);

			assertThat(ozgCommand.getOrder()).isEqualTo(DocumentService.CREATE_ATTACHED_ITEM_ORDER);
		}

		@Test
		void shouldCallVorgangIdMapper() {
			service.buildCreateDocumentOzgCommand(command, BODY_OBJECT);

			verify(commandMapper).toOzgCloudVorgangId(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldSetVorgangId() {
			var expectedVorgangId = OzgCloudVorgangId.from("vorgang-id");
			when(commandMapper.toOzgCloudVorgangId(any())).thenReturn(expectedVorgangId);

			var ozgCommand = buildCreateDocumentOzgCommand();

			assertThat(ozgCommand.getVorgangId()).isEqualTo(expectedVorgangId);
		}

		@Test
		void shouldCallRelationIdMapper() {
			buildCreateDocumentOzgCommand();

			verify(commandMapper).mapRelationId(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldSetRelationId() {
			var expectedRelationId = GenericId.from("relation-id");
			when(commandMapper.mapRelationId(any())).thenReturn(expectedRelationId);

			var ozgCommand = buildCreateDocumentOzgCommand();

			assertThat(ozgCommand.getRelationId()).isEqualTo(expectedRelationId);
		}

		@Test
		void shouldSetBodyObject() {
			var ozgCommand = buildCreateDocumentOzgCommand();

			assertThat(ozgCommand.getBodyObject()).isEqualTo(BODY_OBJECT);
		}

		private OzgCloudCommand buildCreateDocumentOzgCommand() {
			return service.buildCreateDocumentOzgCommand(command, BODY_OBJECT);
		}
	}

	@Nested
	class TestBuildAttachedItem {

		private static final Map<String, Object> ITEM_MAP = Map.of("key", "value");

		@Test
		void shouldSetVorgangId() {
			var attachedItem = buildAttachedItem();

			assertThat(attachedItem).containsEntry(AttachedItem.PROPERTY_VORGANG_ID, CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldSetClient() {
			var attachedItem = buildAttachedItem();

			assertThat(attachedItem).containsEntry(AttachedItem.PROPERTY_CLIENT,
					BescheidCallContextAttachingInterceptor.BESCHEID_MANAGER_CLIENT_NAME);
		}

		@Test
		void shouldSetItemName() {
			var attachedItem = buildAttachedItem();

			assertThat(attachedItem).containsEntry(AttachedItem.PROPERTY_ITEM_NAME, DocumentService.DOCUMENT_ITEM_NAME);
		}

		@Test
		void shouldSetItem() {
			var attachedItem = buildAttachedItem();

			assertThat(attachedItem).containsEntry(AttachedItem.PROPERTY_ITEM, ITEM_MAP);
		}

		private Map<String, Object> buildAttachedItem() {
			return service.buildAttachedItem(command, ITEM_MAP);
		}
	}

	@Nested
	class TestBuildItemMap {

		@Nested
		class TestBuildFromCommand {

			private Command command = CommandTestFactory.createBuilder().relationId(RELATION_ID)
					.bodyObject(Map.of(DocumentService.FIELD_DOCUMENT_FILE, FILE_ID)).build();

			@Test
			void shouldDetDocumentType() {
				var itemMap = buildItemMap();

				assertThat(itemMap).containsEntry(DocumentService.FIELD_DOCUMENT_TYPE, DocumentService.DOCUMENT_TYPE);
			}

			@Test
			void shouldSetDocumentFile() {
				var itemMap = buildItemMap();

				assertThat(itemMap).containsEntry(DocumentService.FIELD_DOCUMENT_FILE, FILE_ID);
			}

			@Test
			void shouldThrowExceptionIfDocumentFileIsMissing() {
				command = CommandTestFactory.createBuilder().bodyObject(Map.of()).build();

				assertThrows(TechnicalException.class, () -> service.buildItemMap(command));
			}

			private Map<String, Object> buildItemMap() {
				return service.buildItemMap(command);
			}
		}

		@Nested
		class TestBuildFromBescheid {

			private BescheidResponse bescheid = BescheidResponseTestFactory.create()
					.withBescheidFileId(FileId.from(DocumentTestFactory.ID));

			@Test
			void shouldDetDocumentType() {
				var itemMap = buildItemMap();

				assertThat(itemMap).containsEntry(DocumentService.FIELD_DOCUMENT_TYPE, DocumentService.DOCUMENT_TYPE);
			}

			@Test
			void shouldSetDocumentFile() {
				var itemMap = buildItemMap();

				assertThat(itemMap).containsEntry(DocumentService.FIELD_DOCUMENT_FILE, DocumentTestFactory.ID);
			}

			@Test
			void shouldSetNachrichtText() {
				var itemMap = buildItemMap();

				assertThat(itemMap).containsEntry(DocumentService.FIELD_NACHRICHT_TEXT, BescheidTestFactory.NACHRICHT_TEXT);
			}

			@Test
			void shouldSetEmptyNachrichtText() {
				bescheid = BescheidResponseTestFactory.createBuilder()
						.bescheidFileId(FileId.from(DocumentTestFactory.ID))
						.nachrichtText(Optional.empty()).build();

				var itemMap = buildItemMap();

				assertThat(itemMap).containsEntry(DocumentService.FIELD_NACHRICHT_TEXT, StringUtils.EMPTY);
			}

			@Test
			void shouldThrowExceptionIfFileIdMissing() {
				bescheid = BescheidResponseTestFactory.createBuilder().bescheidFileId(null).build();

				assertThrows(TechnicalException.class, () -> service.buildItemMap(bescheid));

			}

			private Map<String, Object> buildItemMap() {
				return service.buildItemMap(bescheid);
			}
		}
	}

	@Nested
	class TestGetDocument {

		@Test
		void shouldCallAttachedItemService() {
			service.getDocument(AttachedItemTestFactory.ID);

			verify(attachedItemService).getItem(AttachedItemTestFactory.ID);
		}

		@Test
		void shouldCallDocumentMapper() {
			var expectedItem = AttachedItemTestFactory.createDocument();
			when(attachedItemService.getItem(any())).thenReturn(expectedItem);

			service.getDocument(AttachedItemTestFactory.ID);

			verify(documentMapper).fromAttachedItem(expectedItem);
		}

		@Test
		void shouldReturnDocument() {
			var expectedDocument = DocumentTestFactory.create();
			when(documentMapper.fromAttachedItem(any())).thenReturn(expectedDocument);

			Document document = service.getDocument(AttachedItemTestFactory.ID);

			assertThat(document).isEqualTo(expectedDocument);
		}
	}
}