/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.document.bescheid;

import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.boot.test.mock.mockito.SpyBean;
import org.springframework.context.ApplicationEventPublisher;

import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.common.command.grpc.CommandMapper;
import de.ozgcloud.apilib.user.UserProfileMapper;
import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEventTestFactory;
import de.ozgcloud.command.CommandExecutedEvent;
import de.ozgcloud.command.CommandTestFactory;
import de.ozgcloud.common.test.ITCase;
import de.ozgcloud.document.bescheid.common.callcontext.CurrentUserService;
import de.ozgcloud.document.bescheid.common.user.UserProfile;
import de.ozgcloud.document.bescheid.common.user.UserProfileService;
import de.ozgcloud.document.common.attached_item.AttachedItemService;

@ITCase
class BescheidEventListenerITCase {

	private static final Command COMMAND_WITH_OTHER_ORDER = CommandTestFactory.createBuilder().order("OTHER").build();

	@Autowired
	private ApplicationEventPublisher publisher;
	@SpyBean
	private BescheidEventListener listener;

	@MockBean
	private BescheidService service;
	@MockBean
	private CurrentUserService userService;
	@MockBean
	private UserProfileService userProfileService;
	@MockBean
	private OzgCloudCommandService commandService;
	@MockBean
	private CommandMapper commandMapper;
	@MockBean
	private UserProfileMapper userProfileMapper;
	@MockBean
	private AttachedItemService attachedItemService;

	@Mock
	private UserProfile userProfile;

	@BeforeEach
	void init() {
		when(userProfileService.getUserProfile()).thenReturn(userProfile);
	}

	@Nested
	class TestCreateBescheid {

		private final Command command = CommandTestFactory.createBuilder().order(BescheidEventListener.CREATE_BESCHEID_ORDER).build();

		@Test
		void shouldCallService() {
			publisher.publishEvent(CommandCreatedEventTestFactory.withCommand(command));

			verify(service).createBescheidDraft(command);
		}

		@Test
		void shouldNotReactOnOtherOrder() {
			publisher.publishEvent(CommandCreatedEventTestFactory.withCommand(COMMAND_WITH_OTHER_ORDER));

			verifyNoInteractions(service);
		}
	}

	@Nested
	class TestDeleteBescheid {

		private final Command command = CommandTestFactory.createBuilder().order(BescheidEventListener.DELETE_BESCHEID_ORDER).build();

		@Test
		void shouldCallService() {
			publisher.publishEvent(CommandCreatedEventTestFactory.withCommand(command));

			verify(attachedItemService).deleteBescheidDraft(command);
		}

		@Test
		void shouldNotReactOnOtherOrder() {
			publisher.publishEvent(CommandCreatedEventTestFactory.withCommand(COMMAND_WITH_OTHER_ORDER));

			verifyNoInteractions(attachedItemService);
		}
	}

	@Nested
	class TestUpdateBescheid {

		private final Command command = CommandTestFactory.createBuilder().order(BescheidEventListener.UPDATE_BESCHEID_ORDER).build();

		@Test
		void shouldCallService() {
			publisher.publishEvent(CommandCreatedEventTestFactory.withCommand(command));

			verify(attachedItemService).updateBescheidDraft(any(), any());
		}

		@Test
		void shouldNotReactOnOtherOrder() {
			publisher.publishEvent(CommandCreatedEventTestFactory.withCommand(COMMAND_WITH_OTHER_ORDER));

			verifyNoInteractions(attachedItemService);
		}
	}

	@Nested
	class TestBescheidSent {

		private static final Command COMMAND = CommandTestFactory.createBuilder().order(BescheidEventListener.SEND_BESCHEID_ORDER).build();

		@Test
		void shouldCallMethod() {
			publisher.publishEvent(new CommandExecutedEvent(COMMAND));

			verify(listener).setAntragBewilligung(COMMAND);
		}

		@Test
		void shouldNotReactOnOtherOrder() {
			publisher.publishEvent(new CommandExecutedEvent(COMMAND_WITH_OTHER_ORDER));

			verifyNoInteractions(listener);
		}
	}
}
