/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.document.bescheid;

import static org.assertj.core.api.Assertions.*;
import static org.assertj.core.api.InstanceOfAssertFactories.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.NullSource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.boot.info.BuildProperties;
import org.springframework.test.util.ReflectionTestUtils;

import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.common.command.OzgCloudCreateSubCommandsRequest;
import de.ozgcloud.apilib.common.command.grpc.CommandMapper;
import de.ozgcloud.apilib.common.datatypes.GenericId;
import de.ozgcloud.apilib.user.OzgCloudUserId;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangId;
import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandTestFactory;
import de.ozgcloud.common.binaryfile.FileId;
import de.ozgcloud.common.datatype.StringBasedValue;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.document.DocumentTestFactory;
import de.ozgcloud.document.bescheid.Bescheid.SendBy;
import de.ozgcloud.document.bescheid.administration.AdministrationService;
import de.ozgcloud.document.bescheid.attributes.ClientAttributeService;
import de.ozgcloud.document.bescheid.common.freemarker.TemplateHandler;
import de.ozgcloud.document.bescheid.common.user.UserProfile;
import de.ozgcloud.document.bescheid.common.user.UserProfileService;
import de.ozgcloud.document.bescheid.common.user.UserProfileTestFactory;
import de.ozgcloud.document.bescheid.vorgang.PostfachAddressTestFactory;
import de.ozgcloud.document.bescheid.vorgang.ServiceKontoTestFactory;
import de.ozgcloud.document.bescheid.vorgang.Vorgang;
import de.ozgcloud.document.bescheid.vorgang.VorgangId;
import de.ozgcloud.document.bescheid.vorgang.VorgangService;
import de.ozgcloud.document.bescheid.vorgang.VorgangTestFactory;
import de.ozgcloud.document.common.attached_item.AttachedItem;
import de.ozgcloud.document.common.attached_item.AttachedItemService;
import de.ozgcloud.document.common.attached_item.AttachedItemTestFactory;

class BescheidServiceTest {

	@Spy
	@InjectMocks
	private BescheidService service;

	@Mock
	private VorgangService vorgangService;
	@Mock
	private AttachedItemService attachedItemService;
	@Mock
	private UserProfileService userProfileService;
	@Mock
	private OzgCloudCommandService commandService;
	@Mock
	private CommandMapper commandMapper;
	@Mock
	private ClientAttributeService clientAttributeService;
	@Mock
	private BuildProperties buildProperties;
	@Mock
	private BescheidRemoteService remoteService;
	@Mock
	private BescheidMapper mapper;
	@Mock
	private AdministrationService administrationService;
	@Mock
	private TemplateHandler templateHandler;

	@BeforeEach
	void init() {
		ReflectionTestUtils.setField(service, "remoteService", Optional.of(remoteService));
	}

	@Nested
	class TestFindDraft {

		private final AttachedItem vorgangAttachedItem = AttachedItemTestFactory.create();

		private final Bescheid bescheid = BescheidTestFactory.create();

		@BeforeEach
		void mock() {
			when(attachedItemService.findBescheidItem(any())).thenReturn(Optional.of(vorgangAttachedItem));
			when(mapper.mapFromAttachedItem(any())).thenReturn(bescheid);
		}

		@Test
		void shouldCallVorgangAttachedItemRemoteService() {
			service.findDraft(VorgangTestFactory.ID_STR);

			verify(attachedItemService).findBescheidItem(VorgangTestFactory.ID);
		}

		@Test
		void shouldCallMapper() {
			service.findDraft(VorgangTestFactory.ID_STR);

			verify(mapper).mapFromAttachedItem(vorgangAttachedItem);
		}

		@Test
		void shouldReturnValue() {
			var draft = service.findDraft(VorgangTestFactory.ID_STR);

			assertThat(draft).hasValue(bescheid);
		}
	}

	@DisplayName("Find all")
	@Nested
	class TestFindAll {

		private final AttachedItem attachedItem = AttachedItemTestFactory.create();
		private final Bescheid bescheid = BescheidTestFactory.create();

		@BeforeEach
		void mock() {
			when(attachedItemService.findAllBescheid(any())).thenReturn(Collections.singleton(attachedItem));
			when(mapper.mapFromAttachedItem(any())).thenReturn(bescheid);
		}

		@Test
		void shouldCallVorgangAttachedItemService() {
			findAll();

			verify(attachedItemService).findAllBescheid(VorgangTestFactory.ID);
		}

		@Test
		void shouldCallMapper() {
			findAll();

			verify(mapper).mapFromAttachedItem(attachedItem);
		}

		@Test
		void shouldReturnValue() {
			var bescheide = findAll();

			assertThat(bescheide).containsExactly(bescheid);
		}

		private List<Bescheid> findAll() {
			return service.findAll(VorgangTestFactory.ID).toList();
		}
	}

	@Nested
	class TestCreateBescheidDraft {

		private static final Map<String, Object> COMMAND_BODY = Map.of("key", "value");
		private static final Command COMMAND = CommandTestFactory.createBuilder().bodyObject(COMMAND_BODY).build();
		private static final Bescheid BESCHEID_DRAFT = BescheidTestFactory.create();

		@BeforeEach
		void init() {
			doNothing().when(service).validateBescheidData(any());
			doReturn(BESCHEID_DRAFT).when(service).buildBescheidDraft(any());
		}

		@Test
		void shouldCallValidateBescheidData() {
			service.createBescheidDraft(COMMAND);

			verify(service).validateBescheidData(COMMAND_BODY);
		}

		@Test
		void shouldCallAttachedItemService() {
			service.createBescheidDraft(COMMAND);

			verify(attachedItemService).createBescheidDraft(BESCHEID_DRAFT);
		}

		@Test
		void shouldReturnResult() {
			var createdBescheidId = "created-bescheid-id";
			when(attachedItemService.createBescheidDraft(any(Bescheid.class))).thenReturn(createdBescheidId);

			var result = service.createBescheidDraft(COMMAND);

			assertThat(result).isEqualTo(createdBescheidId);
		}
	}

	@Nested
	class TestValidateBescheidData {

		private static final Command CREATE_BESCHEID_DRAFT_COMMAND = CommandTestFactory.createBuilder()
				.bodyObject(AttachedItemTestFactory.createBescheidItem()).build();

		@Test
		void shouldValidate() {
			assertDoesNotThrow(() -> service.validateBescheidData(CREATE_BESCHEID_DRAFT_COMMAND.getBodyObject()));
		}

		@DisplayName("should fail if bescheidAm is missing")
		@Test
		void shouldThrowExceptionByBescheidAm() {
			var map = Map.<String, Object>of(Bescheid.FIELD_BEWILLIGT, true);

			assertThrows(TechnicalException.class, () -> service.validateBescheidData(map));
		}

		@DisplayName("should fail if bewilligt is missing")
		@Test
		void shouldThrowExceptionByBewilligt() {
			var map = Map.<String, Object>of(Bescheid.FIELD_BESCHIEDEN_AM, "2021-01-01");

			assertThrows(TechnicalException.class, () -> service.validateBescheidData(map));
		}

		@Test
		void shouldValidateSendBy() {
			var bodyObject = Map.<String, Object>of(Bescheid.FIELD_SEND_BY, SendBy.MANUAL.name());
			doNothing().when(service).validateBescheidField(any(), any());

			service.validateBescheidData(bodyObject);

			verify(service).validateSendBy(SendBy.MANUAL.name());
		}

		@DisplayName("validate sendBy")
		@Nested
		class TestValidateSendBy {

			@ValueSource(strings = { "not-exist", "" })
			@ParameterizedTest
			void shouldThrowExceptionOnNonExistingSendByValue(String sendBy) {
				assertThrows(TechnicalException.class, () -> service.validateSendBy(sendBy));
			}

			@EnumSource(names = { "NACHRICHT", "MANUAL" })
			@ParameterizedTest
			void shouldNotThrowAnyExceptionOnExistingSendByValue(SendBy sendBy) {
				assertDoesNotThrow(() -> service.validateSendBy(sendBy.name()));
			}
		}
	}

	@Nested
	class TestBuildBescheidDraft {

		private static final Command CREATE_BESCHEID_DRAFT_COMMAND = CommandTestFactory.createBuilder()
				.bodyObject(AttachedItemTestFactory.createBescheidItem()).build();

		private final Bescheid bescheid = BescheidTestFactory.createBuilder().status(Bescheid.Status.SENT).build();

		@BeforeEach
		void init() {
			when(mapper.mapFromCommand(any())).thenReturn(bescheid);
		}

		@Test
		void shouldCallMapFromCommand() {
			buildBescheidDraft();

			verify(mapper).mapFromCommand(CREATE_BESCHEID_DRAFT_COMMAND);
		}

		@Test
		void shouldResetId() {
			var result = buildBescheidDraft();

			assertThat(result.getId()).isNull();
		}

		@Test
		void shouldResetVersion() {
			var result = buildBescheidDraft();

			assertThat(result.getVersion()).isZero();
		}

		@Test
		void shouldSetStatus() {
			var result = buildBescheidDraft();

			assertThat(result.getStatus()).isEqualTo(Bescheid.Status.DRAFT);
		}

		@Test
		void shouldNotOverrideNachrichtSubject() {
			var result = buildBescheidDraft();

			assertThat(result.getNachrichtSubject()).contains(BescheidTestFactory.NACHRICHT_SUBJECT);
		}

		@Test
		void shouldSetNachrichtSubject() {
			var bescheid = BescheidTestFactory.createBuilder().nachrichtSubject(Optional.empty()).build();
			when(mapper.mapFromCommand(any())).thenReturn(bescheid);

			var result = buildBescheidDraft();

			assertThat(result.getNachrichtSubject()).contains(BescheidService.SUBJECT);
		}

		@Test
		void shouldNotOverrideNachrichtText() {
			var result = buildBescheidDraft();

			assertThat(result.getNachrichtText()).contains(BescheidTestFactory.NACHRICHT_TEXT);
		}

		@Test
		void shouldCallCreateNachrichtText() {
			var createdText = "created text";
			doReturn(createdText).when(service).createNachrichtText();
			var bescheid = BescheidTestFactory.createBuilder().nachrichtText(Optional.empty()).build();
			when(mapper.mapFromCommand(any())).thenReturn(bescheid);

			buildBescheidDraft();

			verify(service).createNachrichtText();
		}

		@Test
		void shouldSetNachrichtText() {
			var createdText = "created text";
			doReturn(createdText).when(service).createNachrichtText();
			var bescheid = BescheidTestFactory.createBuilder().nachrichtText(Optional.empty()).build();
			when(mapper.mapFromCommand(any())).thenReturn(bescheid);

			var result = buildBescheidDraft();

			assertThat(result.getNachrichtText()).contains(createdText);
		}

		private Bescheid buildBescheidDraft() {
			return service.buildBescheidDraft(CREATE_BESCHEID_DRAFT_COMMAND);
		}
	}

	@Nested
	class TestCreateNachrichtText {

		@Captor
		private ArgumentCaptor<Map<String, Object>> signatureCaptor;

		@Test
		void shouldCallGetTemplateParameters() {
			service.createNachrichtText();

			verify(service).getTemplateParameters();
		}

		@Test
		void shouldCallFillTemplate() {
			var parameterMap = Map.<String, Object>of("key", "value");
			doReturn(parameterMap).when(service).getTemplateParameters();

			service.createNachrichtText();

			verify(templateHandler).fillTemplate(BescheidService.DEFAULT_NACHRICHT_TEMPLATE_FILE, parameterMap);
		}

		@Test
		void shouldReturnNachricht() {
			when(templateHandler.fillTemplate(any(), any())).thenReturn(BescheidTestFactory.NACHRICHT_TEXT);

			var result = service.createNachrichtText();

			assertThat(result).isEqualTo(BescheidTestFactory.NACHRICHT_TEXT);
		}
	}

	@Nested
	class TestGetTemplateParameters {

		@Test
		void shouldCallAdministrationService() {
			service.createNachrichtText();

			verify(administrationService).getNachrichtSignature();
		}

		@Test
		void shouldReturnParameters() {
			var signature = "signature";
			when(administrationService.getNachrichtSignature()).thenReturn(Optional.of(signature));

			var result = service.getTemplateParameters();

			assertThat(result).containsEntry(BescheidService.KEY_TEMPLATE_SIGNATURE, signature);
		}

		@Test
		void shouldReturnEmptyMap() {
			var result = service.getTemplateParameters();

			assertThat(result).isEmpty();
		}
	}

	@Nested
	class TestCreateBescheid {

		private final Vorgang vorgang = VorgangTestFactory.create();
		private final BescheidResponse bescheid = BescheidResponseTestFactory.createBuilder().vorgangId(null).build();

		@BeforeEach
		void initMocks() {
			when(vorgangService.getById(any())).thenReturn(vorgang);
			when(remoteService.create(any(), any())).thenReturn(bescheid);
		}

		@Test
		void shouldLoadVorgang() {
			service.createBescheid(BescheidRequestTestFactory.create());

			verify(vorgangService).getById(VorgangTestFactory.ID);
		}

		@Test
		void shouldCreateBescheid() {
			BescheidRequest request = BescheidRequestTestFactory.create();

			service.createBescheid(request);

			verify(remoteService).create(request, vorgang);
		}

		@Test
		void shouldReturnCreatedBescheid() {
			var result = service.createBescheid(BescheidRequestTestFactory.create());

			assertThat(result).usingRecursiveComparison().isEqualTo(BescheidResponseTestFactory.create());
		}
	}

	@Nested
	class TestSendBescheidManually {

		private static final Vorgang VORGANG = VorgangTestFactory.create();

		@Mock
		private OzgCloudCommand subCommand;

		@Captor
		private ArgumentCaptor<VorgangId> vorgangIdCaptor;

		private final AttachedItem bescheidItem = AttachedItemTestFactory.createBescheid();

		@BeforeEach
		void init() {
			doNothing().when(service).validateBescheidSendManually(any(), anyLong());
			doNothing().when(service).addSubCommands(any(), any());
			doReturn(UserProfileTestFactory.ID_STR).when(service).getUserId();
			doReturn(List.of(subCommand)).when(service).buildSetBescheidSentSubCommands(any(), any(), any());
			when(vorgangService.getById(any())).thenReturn(VORGANG);
		}

		@Test
		void shouldCallValidateBescheid() {
			sendBescheid();

			verify(service).validateBescheidSendManually(bescheidItem, AttachedItemTestFactory.VERSION);
		}

		@Test
		void shouldCallGetVorgang() {
			sendBescheid();

			verify(vorgangService).getById(vorgangIdCaptor.capture());
			assertThat(vorgangIdCaptor.getValue()).hasToString(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldCallGetUserId() {
			sendBescheid();

			verify(service).getUserId();
		}

		@Test
		void shouldCallBuildSetBescheidSentSubCommands() {
			sendBescheid();

			verify(service).buildSetBescheidSentSubCommands(bescheidItem, VORGANG, UserProfileTestFactory.ID_STR);
		}

		@Test
		void shouldCallAddSubCommands() {
			sendBescheid();

			verify(service).addSubCommands(CommandTestFactory.ID, List.of(subCommand));
		}

		private void sendBescheid() {
			var parentCommand = CommandTestFactory.createBuilder().relationVersion(AttachedItemTestFactory.VERSION).build();
			service.sendBescheidManually(bescheidItem, parentCommand);
		}
	}

	@Nested
	class TestValidateBescheidSendManually {

		private final AttachedItem attachedItem = createManuallySendBescheid();

		private AttachedItem createManuallySendBescheid() {
			var bescheidItem = AttachedItemTestFactory.createBescheid();
			var item = new HashMap<>(bescheidItem.getItem());

			item.put(Bescheid.FIELD_SEND_BY, SendBy.MANUAL.name());

			return bescheidItem.toBuilder().item(item).build();
		}

		@Test
		void shouldCallValidateBescheid() {
			service.validateBescheidSendManually(attachedItem, AttachedItemTestFactory.VERSION);

			verify(service).validateBescheid(attachedItem, AttachedItemTestFactory.VERSION);
		}

		@Test
		void shouldValidate() {
			doNothing().when(service).validateBescheid(any(), anyLong());

			assertDoesNotThrow(() -> service.validateBescheidSendManually(attachedItem, AttachedItemTestFactory.VERSION));
		}

		@DisplayName("should decline when")
		@ParameterizedTest(name = "sendBy = {0}")
		@EnumSource(value = Bescheid.SendBy.class, names = "MANUAL", mode = EnumSource.Mode.EXCLUDE)
		void shouldDeclineSendByNotManual(Bescheid.SendBy sendBy) {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().itemEntry(Bescheid.FIELD_SEND_BY, sendBy.name()).build();

			var message = assertThrows(TechnicalException.class,
					() -> service.validateBescheidSendManually(bescheidItem, AttachedItemTestFactory.VERSION)).getMessage();
			assertThat(message).contains("sendBy");
		}
	}

	@Nested
	class TestSendBescheidPostfachMail {

		private static final Command PARENT_COMMAND = CommandTestFactory.createBuilder().relationVersion(AttachedItemTestFactory.VERSION).build();
		private static final Vorgang.ServiceKonto SERVICE_KONTO = ServiceKontoTestFactory.create();
		private static final Vorgang VORGANG = VorgangTestFactory.createBuilder().serviceKonto(SERVICE_KONTO).build();

		private static final AttachedItem BESCHEID_ITEM = AttachedItemTestFactory.createBescheid();

		@Mock
		private OzgCloudCommand setBescheidSentSubCommand;
		@Mock
		private OzgCloudCommand sendPostfachNachrichtSubCommand;
		@Captor
		private ArgumentCaptor<VorgangId> vorgangIdCaptor;
		@Captor
		private ArgumentCaptor<List<OzgCloudCommand>> subCommandsCaptor;

		@BeforeEach
		void init() {
			doNothing().when(service).validateBescheidSendPostfach(any(), anyLong());
			doNothing().when(service).addSubCommands(any(), any());
			doReturn(UserProfileTestFactory.ID_STR).when(service).getUserId();
			var subCommands = new ArrayList<OzgCloudCommand>();
			subCommands.add(setBescheidSentSubCommand);
			doReturn(subCommands).when(service).buildSetBescheidSentSubCommands(any(), any(), any());
			doReturn(sendPostfachNachrichtSubCommand).when(service).buildSendPostfachNachrichtCommand(any(), any(), any());
			when(vorgangService.getById(any())).thenReturn(VORGANG);
		}

		@Test
		void shouldCallValidateBescheid() {
			sendBescheid();

			verify(service).validateBescheidSendPostfach(BESCHEID_ITEM, AttachedItemTestFactory.VERSION);
		}

		@Test
		void shouldCallVorgangService() {
			sendBescheid();

			verify(vorgangService).getById(vorgangIdCaptor.capture());
			assertThat(vorgangIdCaptor.getValue()).hasToString(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldCallGetUserId() {
			sendBescheid();

			verify(service).getUserId();
		}

		@Test
		void shouldCallBuildSetBescheidSentSubCommands() {
			sendBescheid();

			verify(service).buildSetBescheidSentSubCommands(BESCHEID_ITEM, VORGANG, UserProfileTestFactory.ID_STR);
		}

		@Test
		void shouldCallBuildSendPostfachNachrichtCommand() {
			sendBescheid();

			verify(service).buildSendPostfachNachrichtCommand(BESCHEID_ITEM, SERVICE_KONTO, UserProfileTestFactory.ID_STR);
		}

		@Test
		void shouldCallAddSubCommands() {
			sendBescheid();

			verify(service).addSubCommands(eq(CommandTestFactory.ID), subCommandsCaptor.capture());
			assertThat(subCommandsCaptor.getValue()).containsExactly(setBescheidSentSubCommand, sendPostfachNachrichtSubCommand);
		}

		private void sendBescheid() {
			service.sendBescheidPostfachMail(BESCHEID_ITEM, PARENT_COMMAND);
		}
	}

	@Nested
	class TestValidateBescheidSendPostfach {

		@Test
		void shouldCallValidateBescheid() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder()
					.itemEntry(Bescheid.FIELD_SEND_BY, Bescheid.SendBy.NACHRICHT.name())
					.itemEntry(Bescheid.FIELD_NACHRICHT_SUBJECT, BescheidTestFactory.NACHRICHT_SUBJECT)
					.itemEntry(Bescheid.FIELD_NACHRICHT_TEXT, BescheidTestFactory.NACHRICHT_TEXT).build();

			service.validateBescheidSendPostfach(bescheidItem, AttachedItemTestFactory.VERSION);

			verify(service).validateBescheid(bescheidItem, AttachedItemTestFactory.VERSION);
		}

		@Test
		void shouldValidate() {
			doNothing().when(service).validateBescheid(any(), anyLong());
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder()
					.itemEntry(Bescheid.FIELD_SEND_BY, Bescheid.SendBy.NACHRICHT.name())
					.itemEntry(Bescheid.FIELD_NACHRICHT_SUBJECT, BescheidTestFactory.NACHRICHT_SUBJECT)
					.itemEntry(Bescheid.FIELD_NACHRICHT_TEXT, BescheidTestFactory.NACHRICHT_TEXT).build();

			assertDoesNotThrow(() -> service.validateBescheidSendPostfach(bescheidItem, AttachedItemTestFactory.VERSION));
		}

		@DisplayName("should decline when")
		@ParameterizedTest(name = "sendBy = {0}")
		@EnumSource(value = Bescheid.SendBy.class, names = "NACHRICHT", mode = EnumSource.Mode.EXCLUDE)
		void shouldDeclineSendByNotPostfach(Bescheid.SendBy sendBy) {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder()
					.itemEntry(Bescheid.FIELD_SEND_BY, sendBy.name())
					.itemEntry(Bescheid.FIELD_NACHRICHT_SUBJECT, BescheidTestFactory.NACHRICHT_SUBJECT)
					.itemEntry(Bescheid.FIELD_NACHRICHT_TEXT, BescheidTestFactory.NACHRICHT_TEXT).build();

			var message = assertThrows(TechnicalException.class,
					() -> service.validateBescheidSendPostfach(bescheidItem, AttachedItemTestFactory.VERSION)).getMessage();
			assertThat(message).contains("sendBy");
		}

		@Test
		void shouldDeclineWhenNoSubject() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder()
					.itemEntry(Bescheid.FIELD_SEND_BY, Bescheid.SendBy.NACHRICHT.name())
					.itemEntry(Bescheid.FIELD_NACHRICHT_SUBJECT, null)
					.itemEntry(Bescheid.FIELD_NACHRICHT_TEXT, BescheidTestFactory.NACHRICHT_TEXT).build();

			var message = assertThrows(TechnicalException.class,
					() -> service.validateBescheidSendPostfach(bescheidItem, AttachedItemTestFactory.VERSION)).getMessage();
			assertThat(message).contains("nachricht subject");
		}

		@Test
		void shouldDeclineWhenNoText() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder()
					.itemEntry(Bescheid.FIELD_SEND_BY, Bescheid.SendBy.NACHRICHT.name())
					.itemEntry(Bescheid.FIELD_NACHRICHT_SUBJECT, BescheidTestFactory.NACHRICHT_SUBJECT)
					.itemEntry(Bescheid.FIELD_NACHRICHT_TEXT, null).build();

			var message = assertThrows(TechnicalException.class,
					() -> service.validateBescheidSendPostfach(bescheidItem, AttachedItemTestFactory.VERSION)).getMessage();
			assertThat(message).contains("nachricht text");
		}
	}

	@Nested
	class TestValidateBescheid {

		private final AttachedItem attachedItem = createManuallySendBescheid();

		private AttachedItem createManuallySendBescheid() {
			var bescheidItem = AttachedItemTestFactory.createBescheid();
			var item = new HashMap<>(bescheidItem.getItem());

			item.put(Bescheid.FIELD_SEND_BY, SendBy.MANUAL.name());

			return bescheidItem.toBuilder().item(item).build();
		}

		@Test
		void shouldValidateBescheid() {
			assertDoesNotThrow(() -> service.validateBescheidSendManually(attachedItem, AttachedItemTestFactory.VERSION));
		}

		@DisplayName("should decline when")
		@ParameterizedTest(name = "status = {0}")
		@EnumSource(value = Bescheid.Status.class, names = "DRAFT", mode = EnumSource.Mode.EXCLUDE)
		void shouldDeclineNotDraft(Bescheid.Status status) {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().itemEntry(Bescheid.FIELD_STATUS, status.name()).build();

			var message = assertThrows(TechnicalException.class,
					() -> service.validateBescheidSendManually(bescheidItem, AttachedItemTestFactory.VERSION)).getMessage();
			assertThat(message).contains("has status");
		}

		@Test
		void shouldDeclineWhenNoDocument() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().itemEntry(Bescheid.FIELD_BESCHEID_DOCUMENT, null).build();

			var message = assertThrows(TechnicalException.class,
					() -> service.validateBescheidSendManually(bescheidItem, AttachedItemTestFactory.VERSION)).getMessage();
			assertThat(message).contains("document");
		}

		@Test
		void shouldDeclineWhenVersionMismatch() {
			var bescheidItem = AttachedItemTestFactory.createBescheid();

			var message = assertThrows(TechnicalException.class,
					() -> service.validateBescheidSendManually(bescheidItem, AttachedItemTestFactory.VERSION + 1)).getMessage();
			assertThat(message).contains("version");
		}
	}

	@Nested
	class TestGetAttachments {

		@Test
		void shouldReturnEmptyList() {
			var result = service.getAttachments(AttachedItemTestFactory.createBescheidBuilder().clearItem().build());

			assertThat(result).isEmpty();
		}

		@Nested
		class TestWithCollection {

			@Test
			void shouldReturnAttachments() {
				var result = service.getAttachments(AttachedItemTestFactory.createBescheid());

				assertThat(result).containsExactly(BescheidTestFactory.ATTACHMENT, BescheidTestFactory.ATTACHMENT);
			}

			@Test
			void shouldReturnEmpty() {
				var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().clearItem()
						.itemEntry(Bescheid.FIELD_ATTACHMENTS, Collections.emptyList()).build();

				var result = service.getAttachments(bescheidItem);

				assertThat(result).isEmpty();
			}

		}

		@Nested
		class TestSingleAttachment {

			private final AttachedItem bescheidItem = AttachedItemTestFactory.createBescheidBuilder().clearItem()
					.itemEntry(Bescheid.FIELD_ATTACHMENTS, BescheidTestFactory.ATTACHMENT).build();

			@Test
			void shouldCallGetStringAsList() {
				service.getAttachments(bescheidItem);

				verify(service).getStringAsList(BescheidTestFactory.ATTACHMENT);
			}

			@Test
			void shouldReturnBescheidFileIdList() {
				var expectedResult = List.of(BescheidTestFactory.ATTACHMENT);
				doReturn(expectedResult).when(service).getStringAsList(any());

				var result = service.getAttachments(bescheidItem);

				assertThat(result).isSameAs(expectedResult);
			}
		}
	}

	@Nested
	class TestGetStringAsList {

		@DisplayName("should return empty list when")
		@ParameterizedTest(name = "value = \"{0}\"")
		@NullSource
		@ValueSource(strings = { "", " " })
		void shouldReturnEmptyListOnBlankValue(String value) {
			var result = service.getStringAsList(value);

			assertThat(result).isEmpty();
		}

		@Test
		void shouldReturnEmptyListOnUnexpectedType() {
			var result = service.getStringAsList(0);

			assertThat(result).isEmpty();
		}
	}

	@Nested
	class TestAddSubCommands {

		@Mock
		private OzgCloudCommand subCommand;
		@Mock
		private OzgCloudCreateSubCommandsRequest createSubCommandsRequest;

		@Test
		void shouldCallBuildCreateSubCommandsRequest() {
			var subCommands = List.of(subCommand);

			service.addSubCommands(CommandTestFactory.ID, subCommands);

			verify(service).buildCreateSubCommandsRequest(CommandTestFactory.ID, subCommands);
		}

		@Test
		void shouldCallCommandService() {
			doReturn(createSubCommandsRequest).when(service).buildCreateSubCommandsRequest(any(), any());

			service.addSubCommands(CommandTestFactory.ID, List.of(subCommand));

			verify(commandService).addSubCommands(createSubCommandsRequest);
		}
	}

	@Nested
	class TestGetBescheidFileId {

		@BeforeEach
		void init() {
			when(attachedItemService.getItem(anyString())).thenReturn(AttachedItemTestFactory.createDocument());
		}

		@Test
		void shouldCallGetItem() {
			service.getBescheidFileId(AttachedItemTestFactory.createBescheid());

			verify(attachedItemService).getItem(DocumentTestFactory.ID);
		}

		@Test
		void shouldReturnBescheidFileId() {
			var result = service.getBescheidFileId(AttachedItemTestFactory.createBescheid());

			assertThat(result).hasToString(DocumentTestFactory.DOCUMENT_FILE);
		}
	}

	@Nested
	class TestGetNachrichtSubject {

		@Test
		void shouldReturnSubject() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().item(Map.of(Bescheid.FIELD_NACHRICHT_SUBJECT,
					BescheidTestFactory.NACHRICHT_SUBJECT)).build();

			var result = service.getNachrichtSubject(bescheidItem);

			assertThat(result).contains(BescheidTestFactory.NACHRICHT_SUBJECT);
		}
	}

	@Nested
	class TestGetNachrichtText {

		@Test
		void shouldReturnText() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().item(Map.of(Bescheid.FIELD_NACHRICHT_TEXT,
					BescheidTestFactory.NACHRICHT_TEXT)).build();

			var result = service.getNachrichtText(bescheidItem);

			assertThat(result).contains(BescheidTestFactory.NACHRICHT_TEXT);
		}
	}

	@Nested
	class TestGetUserId {

		@Mock
		private UserProfile userProfile;

		@BeforeEach
		void init() {
			when(userProfileService.getUserProfile()).thenReturn(userProfile);
			when(userProfile.getId()).thenReturn(UserProfileTestFactory.ID);

		}

		@Test
		void shouldCallGetUserProfile() {
			service.getUserId();

			verify(userProfileService).getUserProfile();
		}

		@Test
		void shouldReturnUserId() {
			var result = service.getUserId();

			assertThat(result).isEqualTo(UserProfileTestFactory.ID_STR);
		}
	}

	@Nested
	class TestBuildSetBescheidSentCommands {

		@Mock
		private AttachedItem bescheidItem;
		@Mock
		private Vorgang vorgang;
		@Mock
		private OzgCloudCommand vorgangBescheidenCommand;
		@Mock
		private OzgCloudCommand setBescheidSentStatusCommand;

		@BeforeEach
		void inti() {
			doReturn(vorgangBescheidenCommand).when(service).buildVorgangBescheidenCommand(any());
			doReturn(setBescheidSentStatusCommand).when(service).buildSetBescheidSentStatusCommand(any(), any());
		}

		@Test
		void shouldCallBuildVorgangBescheidenCommand() {
			buildSetBescheidSentCommands();

			verify(service).buildVorgangBescheidenCommand(vorgang);
		}

		@Test
		void shouldCallBuildSetBescheidSentStatusCommand() {
			buildSetBescheidSentCommands();

			verify(service).buildSetBescheidSentStatusCommand(bescheidItem, UserProfileTestFactory.ID_STR);
		}

		@Test
		void shouldReturnResult() {
			var result = buildSetBescheidSentCommands();

			assertThat(result).containsExactly(vorgangBescheidenCommand, setBescheidSentStatusCommand);
		}

		private List<OzgCloudCommand> buildSetBescheidSentCommands() {
			return service.buildSetBescheidSentSubCommands(bescheidItem, vorgang, UserProfileTestFactory.ID_STR);
		}
	}

	@Nested
	class TestBuildVorgangBescheidenCommand {

		private static final OzgCloudVorgangId VORGANG_ID = OzgCloudVorgangId.from(VorgangTestFactory.ID.toString());
		private static final StringBasedValue RELATION_ID = GenericId.from(VorgangTestFactory.ID.toString());

		@BeforeEach
		void init() {
			when(commandMapper.toOzgCloudVorgangId(anyString())).thenReturn(VORGANG_ID);
			when(commandMapper.mapRelationId(anyString())).thenReturn(RELATION_ID);
		}

		@Test
		void shouldCallMapVorgangId() {
			buildVorgangBescheidenCommand();

			verify(commandMapper).toOzgCloudVorgangId(VorgangTestFactory.ID.toString());
		}

		@Test
		void shouldSetVorgangId() {
			var result = buildVorgangBescheidenCommand();

			assertThat(result.getVorgangId()).isEqualTo(VORGANG_ID);
		}

		@Test
		void shouldCallMapRelationId() {
			buildVorgangBescheidenCommand();

			verify(commandMapper).mapRelationId(VorgangTestFactory.ID.toString());
		}

		@Test
		void shouldSetRelationId() {
			var result = buildVorgangBescheidenCommand();

			assertThat(result.getRelationId()).isEqualTo(RELATION_ID);
		}

		@Test
		void shouldSetRelationVersion() {
			var result = buildVorgangBescheidenCommand();

			assertThat(result.getRelationVersion()).isEqualTo(VorgangTestFactory.VERSION);
		}

		@Test
		void shouldSetOrder() {
			var result = buildVorgangBescheidenCommand();

			assertThat(result.getOrder()).isEqualTo(BescheidService.VORGANG_BESCHEIDEN_ORDER);
		}

		private OzgCloudCommand buildVorgangBescheidenCommand() {
			return service.buildVorgangBescheidenCommand(VorgangTestFactory.create());
		}
	}

	@Nested
	class TestBuildSetBescheidSentStatusCommand {

		private static final StringBasedValue RELATION_ID = GenericId.from(VorgangTestFactory.ID.toString());
		private static final AttachedItem BESCHEID_ITEM = AttachedItemTestFactory.createBescheid();

		@BeforeEach
		void init() {
			when(commandMapper.mapRelationId(any())).thenReturn(RELATION_ID);
		}

		@Test
		void shouldCallMapRelationId() {
			buildSetBescheidSentStatusCommand();

			verify(commandMapper).mapRelationId(AttachedItemTestFactory.ID);
		}

		@Test
		void shouldSetRelationId() {
			var result = buildSetBescheidSentStatusCommand();

			assertThat(result.getRelationId()).isEqualTo(RELATION_ID);
		}

		@Test
		void shouldSetRelationVersion() {
			var result = buildSetBescheidSentStatusCommand();

			assertThat(result.getRelationVersion()).isEqualTo(AttachedItemTestFactory.VERSION);
		}

		@Test
		void shouldSetOrder() {
			var result = buildSetBescheidSentStatusCommand();

			assertThat(result.getOrder()).isEqualTo(AttachedItemService.PATCH_ATTACHED_ITEM);
		}

		@Test
		void shouldSetPropertyId() {
			var result = buildSetBescheidSentStatusCommand();

			assertThat(result.getBodyObject()).containsEntry(AttachedItem.PROPERTY_ID, AttachedItemTestFactory.ID);
		}

		@Test
		void shouldCallBuildBescheidSentStatusItem() {
			buildSetBescheidSentStatusCommand();

			verify(service).buildBescheidSentStatusItem(UserProfileTestFactory.ID_STR);
		}

		@Test
		void shouldSetPropertyItem() {
			var itemMap = Map.of("key", "value");
			doReturn(itemMap).when(service).buildBescheidSentStatusItem(any());

			var result = buildSetBescheidSentStatusCommand();

			assertThat(result.getBodyObject()).containsEntry(AttachedItem.PROPERTY_ITEM, itemMap);
		}

		@Test
		void shouldSetPropertyItemName() {
			var result = buildSetBescheidSentStatusCommand();

			assertThat(result.getBodyObject()).containsEntry(AttachedItem.PROPERTY_ITEM_NAME, AttachedItemService.BESCHEID_ITEM_NAME);
		}

		private OzgCloudCommand buildSetBescheidSentStatusCommand() {
			return service.buildSetBescheidSentStatusCommand(BESCHEID_ITEM, UserProfileTestFactory.ID_STR);
		}
	}

	@Nested
	class TestBuildBescheidSentStatusItem {

		@Test
		void shouldSetStatus() {
			var result = service.buildBescheidSentStatusItem(UserProfileTestFactory.ID_STR);

			assertThat(result).containsEntry(Bescheid.FIELD_STATUS, Bescheid.Status.SENT.name());
		}

		@Test
		void shouldCallBuildSentInfoMap() {
			service.buildBescheidSentStatusItem(UserProfileTestFactory.ID_STR);

			verify(service).buildSentInfoMap(UserProfileTestFactory.ID_STR);
		}

		@Test
		void shouldSetSentInfo() {
			var sentInfoMap = Map.of("key", "value");
			doReturn(sentInfoMap).when(service).buildSentInfoMap(any());

			var result = service.buildBescheidSentStatusItem(UserProfileTestFactory.ID_STR);

			assertThat(result).containsEntry(Bescheid.FIELD_SENT_INFO, sentInfoMap);
		}
	}

	@Nested
	class TestBuildSentInfoMap {

		@Test
		void shouldSetSentAt() {
			var result = service.buildSentInfoMap(UserProfileTestFactory.ID_STR);

			assertThat(getSentAt(result)).isCloseTo(ZonedDateTime.now(), within(1, ChronoUnit.SECONDS));
		}

		@Test
		void shouldSetSentBy() {
			var result = service.buildSentInfoMap(UserProfileTestFactory.ID_STR);

			assertThat(result).containsEntry(SentInfo.FIELD_SENT_BY, UserProfileTestFactory.ID_STR);
		}

		private ZonedDateTime getSentAt(Map<String, Object> sentInfoMap) {
			return ZonedDateTime.parse((String) sentInfoMap.get(SentInfo.FIELD_SENT_AT), DateTimeFormatter.ISO_DATE_TIME);
		}
	}

	@Nested
	class TestBuildSendPostfachNachrichtCommand {

		private static final AttachedItem BESCHEID_ITEM = AttachedItemTestFactory.createBescheid();
		private static final Vorgang.PostfachAddress POSTFACH_ADDRESS = PostfachAddressTestFactory.create();
		private static final Vorgang.ServiceKonto SERVICE_KONTO = ServiceKontoTestFactory.createBuilder().clearPostfachAddresses()
				.postfachAddress(POSTFACH_ADDRESS).build();
		private static final OzgCloudUserId OZG_CLOUD_ID = OzgCloudUserId.from(UserProfileTestFactory.ID_STR);
		private static final Map<String, Object> OBJECT_MAP = Map.of("key", "value");
		private static final OzgCloudVorgangId VORGANG_ID = OzgCloudVorgangId.from(VorgangTestFactory.ID.toString());
		private static final StringBasedValue RELATION_ID = GenericId.from(VorgangTestFactory.ID.toString());

		@BeforeEach
		void init() {
			when(commandMapper.toOzgCloudVorgangId(anyString())).thenReturn(VORGANG_ID);
			when(commandMapper.mapRelationId(anyString())).thenReturn(RELATION_ID);
			when(commandMapper.toOzgCloudUserId(anyString())).thenReturn(OZG_CLOUD_ID);
			doReturn(OBJECT_MAP).when(service).buildSendNachrichtCommandBody(any(), any());
		}

		@Test
		void shouldCallBuildPostfachAddress() {
			buildSendPostfachNachrichtCommand();

			verify(service).buildPostfachAddress(POSTFACH_ADDRESS, ServiceKontoTestFactory.TYPE);
		}

		@Test
		void shouldCallBuildSenNachrichtCommandBody() {
			var postfachAddressMap = Map.<String, Object>of("key", "value");
			doReturn(postfachAddressMap).when(service).buildPostfachAddress(any(), any());

			buildSendPostfachNachrichtCommand();

			verify(service).buildSendNachrichtCommandBody(BESCHEID_ITEM, postfachAddressMap);
		}

		@Test
		void shouldSetBodyObject() {
			var result = buildSendPostfachNachrichtCommand();

			assertThat(result.getBodyObject()).isSameAs(OBJECT_MAP);
		}

		@Test
		void shouldCallMapVorgangId() {
			buildSendPostfachNachrichtCommand();

			verify(commandMapper).toOzgCloudVorgangId(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldSetVorgangId() {
			var result = buildSendPostfachNachrichtCommand();

			assertThat(result.getVorgangId()).isEqualTo(VORGANG_ID);
		}

		@Test
		void shouldCallMapRelationId() {
			buildSendPostfachNachrichtCommand();

			verify(commandMapper).mapRelationId(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldSetRelationId() {
			var result = buildSendPostfachNachrichtCommand();

			assertThat(result.getRelationId()).isEqualTo(RELATION_ID);
		}

		@Test
		void shouldSetOrder() {
			var result = buildSendPostfachNachrichtCommand();

			assertThat(result.getOrder()).isEqualTo(BescheidService.SEND_POSTFACH_NACHRICHT_ORDER);
		}

		@Test
		void shouldCallMapToOzgCloudUserId() {
			buildSendPostfachNachrichtCommand();

			verify(commandMapper).toOzgCloudUserId(UserProfileTestFactory.ID_STR);
		}

		@Test
		void shouldSetCreatedBy() {
			var result = buildSendPostfachNachrichtCommand();

			assertThat(result.getCreatedBy()).isEqualTo(OZG_CLOUD_ID);
		}

		private OzgCloudCommand buildSendPostfachNachrichtCommand() {
			return service.buildSendPostfachNachrichtCommand(BESCHEID_ITEM, SERVICE_KONTO, UserProfileTestFactory.ID_STR);
		}
	}

	@Nested
	class TestBuildPostfachAddress {

		@Test
		void shouldSetType() {
			var result = buildPostfachAddress();

			assertThat(result).containsEntry(Vorgang.PostfachAddress.FIELD_TYPE, PostfachAddressTestFactory.TYPE);
		}

		@Test
		void shouldSetVersion() {
			var result = buildPostfachAddress();

			assertThat(result).containsEntry(Vorgang.PostfachAddress.FIELD_VERSION, PostfachAddressTestFactory.VERSION);
		}

		@Test
		void shouldSetIdentifier() {
			var result = buildPostfachAddress();

			assertThat(result).extracting(Vorgang.PostfachAddress.FIELD_IDENTIFIER, MAP)
					.containsEntry(BescheidService.FIELD_POSTFACH_ID, PostfachAddressTestFactory.POSTFACH_ID);
		}

		@Test
		void shouldSetServicekontoType() {
			var result = buildPostfachAddress();

			assertThat(result).containsEntry(Vorgang.ServiceKonto.FIELD_SERVICEKONTO_TYPE, ServiceKontoTestFactory.TYPE);
		}

		private Map<String, Object> buildPostfachAddress() {
			return service.buildPostfachAddress(PostfachAddressTestFactory.create(), ServiceKontoTestFactory.TYPE);
		}
	}

	@Nested
	class TestBuildSendNachrichtCommandBody {

		private static final Map<String, Object> SERVICE_KONTO_MAP = Map.of("key", "value");

		@Mock
		private AttachedItem bescheidItem;

		private Map<String, Object> buildSendNachrichtCommandBody() {
			return service.buildSendNachrichtCommandBody(bescheidItem, SERVICE_KONTO_MAP);
		}

		@Nested
		class TestWithValuesFromBescheidItem {

			@BeforeEach
			void init() {
				doReturn(Optional.of(BescheidTestFactory.NACHRICHT_SUBJECT)).when(service).getNachrichtSubject(any());
				doReturn(Optional.of(BescheidTestFactory.NACHRICHT_TEXT)).when(service).getNachrichtText(any());
				doReturn(List.of(BescheidTestFactory.ATTACHMENT)).when(service).buildAttachments(any());
			}

			@Test
			void shouldSetReplyOption() {
				var result = buildSendNachrichtCommandBody();

				assertThat(result).containsEntry(BescheidService.FIELD_REPLY_OPTION, BescheidService.REPLY_OPTION);
			}

			@Test
			void shouldCallGetNachrichtSubject() {
				buildSendNachrichtCommandBody();

				verify(service).getNachrichtSubject(bescheidItem);
			}

			@Test
			void shouldSetSubject() {
				var result = buildSendNachrichtCommandBody();

				assertThat(result).containsEntry(BescheidService.FIELD_SUBJECT, BescheidTestFactory.NACHRICHT_SUBJECT);
			}

			@Test
			void shouldCallGetNachrichtText() {
				buildSendNachrichtCommandBody();

				verify(service).getNachrichtText(bescheidItem);
			}

			@Test
			void shouldSetMailBody() {
				var result = buildSendNachrichtCommandBody();

				assertThat(result).containsEntry(BescheidService.FIELD_MAIL_BODY, BescheidTestFactory.NACHRICHT_TEXT);
			}

			@Test
			void shouldCallBuildAttachments() {
				buildSendNachrichtCommandBody();

				verify(service).buildAttachments(bescheidItem);
			}

			@Test
			void shouldSetAttachments() {
				var result = buildSendNachrichtCommandBody();

				assertThat(result).extracting(BescheidService.FIELD_ATTACHMENTS, LIST).containsExactly(BescheidTestFactory.ATTACHMENT);
			}

			@Test
			void shouldSetPostfachAddress() {
				var result = buildSendNachrichtCommandBody();

				assertThat(result).containsEntry(Vorgang.ServiceKonto.FIELD_POSTFACH_ADDRESS, SERVICE_KONTO_MAP);
			}
		}

		@Nested
		class TestWithDefaultValues {

			@BeforeEach
			void init() {
				doReturn(Optional.empty()).when(service).getNachrichtSubject(any());
				doReturn(Optional.empty()).when(service).getNachrichtText(any());
				doReturn(Collections.emptyList()).when(service).buildAttachments(any());
			}

			@Test
			void shouldSetSubject() {
				var result = buildSendNachrichtCommandBody();

				assertThat(result).containsEntry(BescheidService.FIELD_SUBJECT, BescheidService.SUBJECT);
			}

			@Test
			void shouldSetMailBody() {
				var result = buildSendNachrichtCommandBody();

				assertThat(result).containsEntry(BescheidService.FIELD_MAIL_BODY, StringUtils.EMPTY);
			}
		}
	}

	@Nested
	class TestBuildAttachments {

		private static final String BESCHEID_FILE_ID = "bescheid-file-id";
		private static final String ATTACHMENT_FILE_ID = "attachment-file-id";

		@Mock
		private AttachedItem bescheidItem;

		@BeforeEach
		void init() {
			doReturn(FileId.from(BESCHEID_FILE_ID)).when(service).getBescheidFileId(any());
			doReturn(List.of(FileId.from(ATTACHMENT_FILE_ID))).when(service).getAttachments(any());
		}

		@Test
		void shouldCallGetBescheidFileId() {
			buildAttachments();

			verify(service).getBescheidFileId(bescheidItem);
		}

		@Test
		void shouldCallGetAttachments() {
			buildAttachments();

			verify(service).getAttachments(bescheidItem);
		}

		@Test
		void shouldReturnAllAttachments() {
			var result = buildAttachments();

			assertThat(result).containsExactly(BESCHEID_FILE_ID, ATTACHMENT_FILE_ID);
		}

		private List<String> buildAttachments() {
			return service.buildAttachments(bescheidItem);
		}
	}

	@Nested
	class TestSetAntragResult {

		private final AttachedItem bescheidItem = AttachedItemTestFactory.createBescheid();

		@Test
		void shouldCallGetBewilligt() {
			service.setAntragBewilligung(bescheidItem);

			verify(service).getBewilligt(bescheidItem);
		}

		@Test
		void shouldCallClientAttributeService() {
			doReturn(true).when(service).getBewilligt(any());

			service.setAntragBewilligung(bescheidItem);

			verify(clientAttributeService).setAntragResult(CommandTestFactory.VORGANG_ID, true);
		}

		@Test
		void shouldCatchException() {
			doThrow(new TechnicalException("error")).when(clientAttributeService).setAntragResult(anyString(), anyBoolean());

			assertDoesNotThrow(() -> service.setAntragBewilligung(bescheidItem));
		}
	}

	@Nested
	class TestGetBewilligt {

		@Test
		void shouldReturnBewilligt() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().itemEntry(Bescheid.FIELD_BEWILLIGT, true).build();

			var result = service.getBewilligt(bescheidItem);

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnDefault() {
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().clearItem().build();

			var result = service.getBewilligt(bescheidItem);

			assertThat(result).isFalse();
		}
	}

	@Nested
	class TestGetConfig {

		@Test
		void shouldCallGetVersion() {
			service.getConfig();

			verify(buildProperties).getVersion();
		}

		@Test
		void shouldSetBescheidManagerVersion() {
			when(buildProperties.getVersion()).thenReturn(BescheidManagerConfigTestFactory.VERSION);

			var result = service.getConfig();

			assertThat(result.getVersion()).isEqualTo(BescheidManagerConfigTestFactory.VERSION);
		}

		@Test
		void shouldCallGetJavaVersion() {
			service.getConfig();

			verify(service).getJavaVersion();
		}

		@Test
		void shouldSetJavaVersion() {
			when(service.getJavaVersion()).thenReturn(BescheidManagerConfigTestFactory.JAVA_VERSION);

			var result = service.getConfig();

			assertThat(result.getJavaVersion()).isEqualTo(BescheidManagerConfigTestFactory.JAVA_VERSION);
		}

		@Test
		void shouldCallBuildFeatures() {
			service.getConfig();

			verify(service).buildFeatures();
		}

		@Test
		void shouldSetFeatures() {
			var features = BescheidManagerConfig.Features.builder().build();
			when(service.buildFeatures()).thenReturn(features);

			var result = service.getConfig();

			assertThat(result.getFeatures()).isSameAs(features);
		}

	}

	@Nested
	class TestBuildFeatures {

		@Test
		void shouldSetCannotCreateBescheidDocument() {
			ReflectionTestUtils.setField(service, "remoteService", Optional.empty());

			var result = service.buildFeatures();

			assertThat(result.isCanCreateBescheidDocument()).isFalse();
		}

		@Test
		void shouldSetCanCreateBescheidDocument() {
			var result = service.buildFeatures();

			assertThat(result.isCanCreateBescheidDocument()).isTrue();
		}

	}

	@DisplayName("Get bescheid")
	@Nested
	class TestGetBescheid {

		private final AttachedItem attachedItem = AttachedItemTestFactory.create();

		@BeforeEach
		void mock() {
			when(attachedItemService.getItem(any())).thenReturn(attachedItem);
			when(mapper.mapFromAttachedItem(any())).thenReturn(BescheidTestFactory.create());
		}

		@Test
		void shouldCallVorgangAttachedItemService() {
			service.getBescheid(AttachedItemTestFactory.ID);

			verify(attachedItemService).getItem(AttachedItemTestFactory.ID);
		}

		@Test
		void shouldCallMapper() {
			service.getBescheid(AttachedItemTestFactory.ID);

			verify(mapper).mapFromAttachedItem(attachedItem);
		}

		@Test
		void shouldReturnValue() {
			var bescheid = service.getBescheid(AttachedItemTestFactory.ID);

			assertThat(bescheid).usingRecursiveComparison().isEqualTo(BescheidTestFactory.create());
		}
	}
}
