/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.document.common.attached_item;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.document.bescheid.Bescheid;
import de.ozgcloud.document.bescheid.BescheidCallContextAttachingInterceptor;
import de.ozgcloud.document.bescheid.vorgang.VorgangTestFactory;
import de.ozgcloud.command.CommandTestFactory;
import de.ozgcloud.vorgang.common.GrpcObject;
import de.ozgcloud.vorgang.common.GrpcProperty;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcFindVorgangAttachedItemRequest;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcFindVorgangAttachedItemResponse;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItem;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItemRequest;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItemResponse;
import de.ozgcloud.vorgang.vorgangAttachedItem.VorgangAttachedItemServiceGrpc.VorgangAttachedItemServiceBlockingStub;
import io.grpc.ClientInterceptor;

class VorgangAttachedItemRemoteServiceTest {

	@Spy
	@InjectMocks
	private VorgangAttachedItemRemoteService service;

	@Mock
	private VorgangAttachedItemServiceBlockingStub serviceStub;
	@Mock
	private ClientInterceptor bescheidCallContextInterceptor;
	@Mock
	private AttachedItemMapper attachedItemMapper;

	@Nested
	class TestFindBescheidDraft {

		@Mock
		private GrpcFindVorgangAttachedItemRequest request;

		@Nested
		class TestFindByVorgangId {

			private static final AttachedItem BESCHEID_ITEM = AttachedItemTestFactory.createBescheid();

			@BeforeEach
			void init() {
				doReturn(request).when(service).buildFindRequest(any());
				doReturn(Optional.of(BESCHEID_ITEM)).when(service).findBescheidDraft(any(GrpcFindVorgangAttachedItemRequest.class));
			}

			@Test
			void shouldCallCreateFindRequestBuilder() {
				service.findBescheidDraft(VorgangTestFactory.ID);

				verify(service).buildFindRequest(VorgangTestFactory.ID_STR);
			}

			@Test
			void shouldCallGetBescheidDraft() {
				service.findBescheidDraft(VorgangTestFactory.ID);

				verify(service).findBescheidDraft(request);
			}

			@Test
			void shouldReturnResult() {
				var result = service.findBescheidDraft(VorgangTestFactory.ID);

				assertThat(result).contains(BESCHEID_ITEM);
			}
		}

		@Nested
		@DisplayName("by grpc request")
		class TestByGrpcRequest {

			private static final AttachedItem BESCHEID_ITEM = AttachedItemTestFactory.createBescheid();

			@Mock
			private GrpcFindVorgangAttachedItemResponse response;

			@BeforeEach
			void init() {
			}

			@Test
			void shouldCallFindAllBescheid() {
				doReturn(Stream.of(BESCHEID_ITEM)).when(service).findAllBescheid(any(GrpcFindVorgangAttachedItemRequest.class));

				service.findBescheidDraft(request);

				verify(service).findAllBescheid(request);
			}

			@Test
			void shouldCallHasStatusDraft() {
				doReturn(Stream.of(BESCHEID_ITEM)).when(service).findAllBescheid(any(GrpcFindVorgangAttachedItemRequest.class));

				service.findBescheidDraft(request);

				verify(service).hasStatusDraft(BESCHEID_ITEM);
			}

			@Test
			void shouldFilterDraftItems() {
				doReturn(Stream.of(BESCHEID_ITEM)).when(service).findAllBescheid(any(GrpcFindVorgangAttachedItemRequest.class));
				doReturn(true).when(service).hasStatusDraft(any());

				var result = service.findBescheidDraft(request);

				assertThat(result).contains(BESCHEID_ITEM);
			}

			@Test
			void shouldReturnEmptyWhenNotFound() {
				doReturn(Stream.empty()).when(service).findAllBescheid(any(GrpcFindVorgangAttachedItemRequest.class));

				var result = service.findBescheidDraft(request);

				assertThat(result).isEmpty();
			}

			@Test
			void shouldReturnEmptyWhenNotDraft() {
				doReturn(Stream.of(BESCHEID_ITEM)).when(service).findAllBescheid(any(GrpcFindVorgangAttachedItemRequest.class));
				doReturn(false).when(service).hasStatusDraft(any());

				var result = service.findBescheidDraft(request);

				assertThat(result).isEmpty();
			}

		}

	}

	@Nested
	class TestFindAllBescheid {

		private static final AttachedItem RESULT_BESCHEID_ITEM = AttachedItemTestFactory.createBescheid();

		@Mock
		private GrpcFindVorgangAttachedItemRequest request;

		@BeforeEach
		void init() {
			doReturn(Stream.of(RESULT_BESCHEID_ITEM)).when(service).findAllBescheid(any(GrpcFindVorgangAttachedItemRequest.class));
		}

		@Test
		void shouldCallBuildFindRequest() {
			service.findAllBescheid(VorgangTestFactory.ID);

			verify(service).buildFindRequest(VorgangTestFactory.ID.toString());
		}

		@Test
		void shouldCallFindAllBescheid() {
			doReturn(request).when(service).buildFindRequest(any());

			service.findAllBescheid(VorgangTestFactory.ID);

			verify(service).findAllBescheid(request);
		}

		@Test
		void shouldReturnResult() {
			var result = service.findAllBescheid(VorgangTestFactory.ID);

			assertThat(result).containsExactly(RESULT_BESCHEID_ITEM);
		}
	}

	@Nested
	class TestBuildFindRequest {

		@Test
		void shouldBuildRequest() {
			var request = buildFindRequest();

			assertThat(request.getVorgangId()).isEqualTo(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldSetItemName() {
			var request = buildFindRequest();

			assertThat(request.getItemName()).isEqualTo(AttachedItemService.BESCHEID_ITEM_NAME);
		}

		@Test
		void shouldSetClient() {
			var request = buildFindRequest();

			assertThat(request.getClient()).isEqualTo(BescheidCallContextAttachingInterceptor.BESCHEID_MANAGER_CLIENT_NAME);
		}

		private GrpcFindVorgangAttachedItemRequest buildFindRequest() {
			return service.buildFindRequest(CommandTestFactory.VORGANG_ID);
		}
	}

	@Nested
	class TestFindAllBescheidWithGrpcRequest {

		@Mock
		private GrpcFindVorgangAttachedItemRequest request;
		@Mock
		private GrpcFindVorgangAttachedItemResponse response;
		@Mock
		private GrpcVorgangAttachedItem grpcVorgangAttachedItem;

		@BeforeEach
		void init() {
			doReturn(serviceStub).when(service).getServiceStub();
			when(serviceStub.find(any())).thenReturn(response);
		}

		@Test
		void shouldCallGetServiceStub() {
			service.findAllBescheid(request);

			verify(service).getServiceStub();
		}

		@Test
		void shouldCallMapper() {
			when(response.getVorgangAttachedItemsList()).thenReturn(Collections.singletonList(grpcVorgangAttachedItem));

			service.findAllBescheid(request).toList();

			verify(attachedItemMapper).mapFromVorgangAttachedItem(grpcVorgangAttachedItem);
		}

		@Test
		void shouldReturnSortedResult() {
			var beschiedenAmFirst = "2021-01-01";
			var beschiedenAmLast = "2024-01-01";
			createItems(beschiedenAmFirst, beschiedenAmLast);

			var result = service.findAllBescheid(request).toList();

			assertThat(result).hasSize(2);
			assertThat(result.get(0).getItem()).containsEntry(Bescheid.FIELD_BESCHIEDEN_AM, beschiedenAmLast);
			assertThat(result.get(1).getItem()).containsEntry(Bescheid.FIELD_BESCHIEDEN_AM, beschiedenAmFirst);
		}

		private void createItems(String beschiedenAmFirst, String beschiedenAmLast) {
			var itemBeschiedenFirst = createGrpcAttachedItem(beschiedenAmLast);
			when(attachedItemMapper.mapFromVorgangAttachedItem(itemBeschiedenFirst)).thenReturn(createAttachedItem(beschiedenAmLast));
			var itemBeschiedenLast = createGrpcAttachedItem(beschiedenAmFirst);
			when(attachedItemMapper.mapFromVorgangAttachedItem(itemBeschiedenLast)).thenReturn(createAttachedItem(beschiedenAmFirst));
			when(response.getVorgangAttachedItemsList()).thenReturn(List.of(itemBeschiedenFirst, itemBeschiedenLast));
		}

		private GrpcVorgangAttachedItem createGrpcAttachedItem(String beschiedenAm) {
			return GrpcVorgangAttachedItemTestFactory.createBuilder().clearItem().setItem(GrpcObject.newBuilder().addProperty(
					GrpcProperty.newBuilder().setName(Bescheid.FIELD_BESCHIEDEN_AM).addValue(beschiedenAm))).build();
		}

		private AttachedItem createAttachedItem(String beschiedenAm) {
			return AttachedItemTestFactory.createBescheidBuilder().clearItem().itemEntry(Bescheid.FIELD_BESCHIEDEN_AM, beschiedenAm)
					.build();
		}
	}

	@Nested
	class TestHasStatusDraft {

		@Test
		void shouldReturnTrueIfDraft() {
			var item = AttachedItemTestFactory.createBescheidBuilder().clearItem()
					.itemEntry(Bescheid.FIELD_STATUS, Bescheid.Status.DRAFT.name())
					.build();

			var result = service.hasStatusDraft(item);

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalseIfNotDraft() {
			var item = AttachedItemTestFactory.createBescheidBuilder().clearItem()
					.itemEntry(Bescheid.FIELD_STATUS, Bescheid.Status.SENT.name())
					.build();

			var result = service.hasStatusDraft(item);

			assertThat(result).isFalse();
		}
	}

	@Nested
	class TestGetBescheid {

		@Mock
		private GrpcVorgangAttachedItemRequest grpcVorgangAttachedItemRequest;
		@Mock
		private GrpcVorgangAttachedItemResponse grpcVorgangAttachedItemResponse;
		@Mock
		private GrpcVorgangAttachedItem grpcVorgangAttachedItem;

		@BeforeEach
		void init() {
			when(serviceStub.getById(any())).thenReturn(grpcVorgangAttachedItemResponse);
			doReturn(serviceStub).when(service).getServiceStub();
		}

		@Test
		void shouldCallGetServiceStab() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			getItem();

			verify(service).getServiceStub();
		}

		@Test
		void shouldCallBuildRequest() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			getItem();

			verify(service).buildGetByIdRequest(AttachedItemTestFactory.ID);
		}

		@Test
		void shouldCallGetById() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);
			doReturn(grpcVorgangAttachedItemRequest).when(service).buildGetByIdRequest(any());

			getItem();

			verify(serviceStub).getById(grpcVorgangAttachedItemRequest);
		}

		@Test
		void shouldCallMapper() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			getItem();

			verify(attachedItemMapper).mapFromVorgangAttachedItem(grpcVorgangAttachedItem);
		}

		@Test
		void shouldReturnFoundBescheid() {
			var expectedBescheid = AttachedItemTestFactory.createBescheid();
			when(attachedItemMapper.mapFromVorgangAttachedItem(any())).thenReturn(expectedBescheid);
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			var result = getItem();

			assertThat(result).isEqualTo(expectedBescheid);
		}

		private AttachedItem getItem() {
			return service.getItem(AttachedItemTestFactory.ID);
		}
	}

	@Nested
	class TestGetItem {

		@Mock
		private GrpcVorgangAttachedItemRequest grpcVorgangAttachedItemRequest;
		@Mock
		private GrpcVorgangAttachedItemResponse grpcVorgangAttachedItemResponse;
		@Mock
		private GrpcVorgangAttachedItem grpcVorgangAttachedItem;

		@BeforeEach
		void init() {
			when(serviceStub.getById(any())).thenReturn(grpcVorgangAttachedItemResponse);
			doReturn(serviceStub).when(service).getServiceStub();
		}

		@Test
		void shouldCallGetServiceStab() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			getItem();

			verify(service).getServiceStub();
		}

		@Test
		void shouldCallBuildRequest() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			getItem();

			verify(service).buildGetByIdRequest(AttachedItemTestFactory.ID);
		}

		@Test
		void shouldCallGetById() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);
			doReturn(grpcVorgangAttachedItemRequest).when(service).buildGetByIdRequest(any());

			getItem();

			verify(serviceStub).getById(grpcVorgangAttachedItemRequest);
		}

		@Test
		void shouldCallMapper() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			getItem();

			verify(attachedItemMapper).mapFromVorgangAttachedItem(grpcVorgangAttachedItem);
		}

		@Test
		void shouldReturnFoundBescheid() {
			var expectedItem = AttachedItemTestFactory.createDocument();
			when(attachedItemMapper.mapFromVorgangAttachedItem(any())).thenReturn(expectedItem);
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			var result = getItem();

			assertThat(result).isEqualTo(expectedItem);
		}

		private AttachedItem getItem() {
			return service.getItem(AttachedItemTestFactory.ID);
		}
	}

	@Nested
	class TestBuildGetByIdRequest {

		@Test
		void shouldSetBescheidId() {
			var request = service.buildGetByIdRequest(AttachedItemTestFactory.ID);

			assertThat(request.getId()).isEqualTo(AttachedItemTestFactory.ID);
		}
	}

	@Nested
	class TestGetServiceStub {

		@Test
		void shouldCallWithInterceptors() {
			service.getServiceStub();

			verify(serviceStub).withInterceptors(bescheidCallContextInterceptor);
		}
	}
}